@@ -25,8 +25,9 @@ from os_client_config.tests import base
 fake_config_dict = {'a': 1, 'os_b': 2, 'c': 3, 'os_c': 4}
 fake_services_dict = {
     'compute_api_version': '2',
-    'compute_endpoint': 'http://compute.example.com',
+    'compute_endpoint_override': 'http://compute.example.com',
     'compute_region_name': 'region-bl',
+    'telemetry_endpoint': 'http://telemetry.example.com',
     'interface': 'public',
     'image_service_type': 'mage',
     'identity_interface': 'admin',
@@ -47,7 +48,7 @@ class TestCloudConfig(base.TestCase):
         self.assertEqual(1, cc.a)
 
         # Look up prefixed attribute, fail - returns None
-        self.assertEqual(None, cc.os_b)
+        self.assertIsNone(cc.os_b)
 
         # Look up straight value, then prefixed value
         self.assertEqual(3, cc.c)
@@ -139,7 +140,7 @@ class TestCloudConfig(base.TestCase):
         self.assertEqual('region-al', cc.get_region_name())
         self.assertEqual('region-al', cc.get_region_name('image'))
         self.assertEqual('region-bl', cc.get_region_name('compute'))
-        self.assertEqual(None, cc.get_api_version('image'))
+        self.assertIsNone(cc.get_api_version('image'))
         self.assertEqual('2', cc.get_api_version('compute'))
         self.assertEqual('mage', cc.get_service_type('image'))
         self.assertEqual('compute', cc.get_service_type('compute'))
@@ -149,7 +150,7 @@ class TestCloudConfig(base.TestCase):
                          cc.get_endpoint('compute'))
         self.assertEqual(None,
                          cc.get_endpoint('image'))
-        self.assertEqual(None, cc.get_service_name('compute'))
+        self.assertIsNone(cc.get_service_name('compute'))
         self.assertEqual('locks', cc.get_service_name('identity'))
 
     def test_volume_override(self):
@@ -189,14 +190,24 @@ class TestCloudConfig(base.TestCase):
             verify=True, cert=None, timeout=9)
 
     @mock.patch.object(ksa_session, 'Session')
-    def test_override_session_endpoint(self, mock_session):
+    def test_override_session_endpoint_override(self, mock_session):
         config_dict = defaults.get_defaults()
         config_dict.update(fake_services_dict)
         cc = cloud_config.CloudConfig(
             "test1", "region-al", config_dict, auth_plugin=mock.Mock())
         self.assertEqual(
             cc.get_session_endpoint('compute'),
-            fake_services_dict['compute_endpoint'])
+            fake_services_dict['compute_endpoint_override'])
+
+    @mock.patch.object(ksa_session, 'Session')
+    def test_override_session_endpoint(self, mock_session):
+        config_dict = defaults.get_defaults()
+        config_dict.update(fake_services_dict)
+        cc = cloud_config.CloudConfig(
+            "test1", "region-al", config_dict, auth_plugin=mock.Mock())
+        self.assertEqual(
+            cc.get_session_endpoint('telemetry'),
+            fake_services_dict['telemetry_endpoint'])
 
     @mock.patch.object(cloud_config.CloudConfig, 'get_session')
     def test_session_endpoint_identity(self, mock_get_session):
@@ -294,9 +305,96 @@ class TestCloudConfig(base.TestCase):
             "test1", "region-al", config_dict, auth_plugin=mock.Mock())
         cc.get_legacy_client('image', mock_client)
         mock_client.assert_called_with(
-            '2',
+            2.0,
             service_name=None,
-            endpoint='http://example.com',
+            endpoint_override='http://example.com',
+            region_name='region-al',
+            interface='public',
+            session=mock.ANY,
+            # Not a typo - the config dict above overrides this
+            service_type='mage'
+        )
+
+    @mock.patch.object(cloud_config.CloudConfig, 'get_session_endpoint')
+    def test_legacy_client_image_override(self, mock_get_session_endpoint):
+        mock_client = mock.Mock()
+        mock_get_session_endpoint.return_value = 'http://example.com/v2'
+        config_dict = defaults.get_defaults()
+        config_dict.update(fake_services_dict)
+        config_dict['image_endpoint_override'] = 'http://example.com/override'
+        cc = cloud_config.CloudConfig(
+            "test1", "region-al", config_dict, auth_plugin=mock.Mock())
+        cc.get_legacy_client('image', mock_client)
+        mock_client.assert_called_with(
+            2.0,
+            service_name=None,
+            endpoint_override='http://example.com/override',
+            region_name='region-al',
+            interface='public',
+            session=mock.ANY,
+            # Not a typo - the config dict above overrides this
+            service_type='mage'
+        )
+
+    @mock.patch.object(cloud_config.CloudConfig, 'get_session_endpoint')
+    def test_legacy_client_image_versioned(self, mock_get_session_endpoint):
+        mock_client = mock.Mock()
+        mock_get_session_endpoint.return_value = 'http://example.com/v2'
+        config_dict = defaults.get_defaults()
+        config_dict.update(fake_services_dict)
+        # v2 endpoint was passed, 1 requested in config, endpoint wins
+        config_dict['image_api_version'] = '1'
+        cc = cloud_config.CloudConfig(
+            "test1", "region-al", config_dict, auth_plugin=mock.Mock())
+        cc.get_legacy_client('image', mock_client)
+        mock_client.assert_called_with(
+            2.0,
+            service_name=None,
+            endpoint_override='http://example.com',
+            region_name='region-al',
+            interface='public',
+            session=mock.ANY,
+            # Not a typo - the config dict above overrides this
+            service_type='mage'
+        )
+
+    @mock.patch.object(cloud_config.CloudConfig, 'get_session_endpoint')
+    def test_legacy_client_image_unversioned(self, mock_get_session_endpoint):
+        mock_client = mock.Mock()
+        mock_get_session_endpoint.return_value = 'http://example.com/'
+        config_dict = defaults.get_defaults()
+        config_dict.update(fake_services_dict)
+        # Versionless endpoint, config wins
+        config_dict['image_api_version'] = '1'
+        cc = cloud_config.CloudConfig(
+            "test1", "region-al", config_dict, auth_plugin=mock.Mock())
+        cc.get_legacy_client('image', mock_client)
+        mock_client.assert_called_with(
+            '1',
+            service_name=None,
+            endpoint_override='http://example.com',
+            region_name='region-al',
+            interface='public',
+            session=mock.ANY,
+            # Not a typo - the config dict above overrides this
+            service_type='mage'
+        )
+
+    @mock.patch.object(cloud_config.CloudConfig, 'get_session_endpoint')
+    def test_legacy_client_image_argument(self, mock_get_session_endpoint):
+        mock_client = mock.Mock()
+        mock_get_session_endpoint.return_value = 'http://example.com/v3'
+        config_dict = defaults.get_defaults()
+        config_dict.update(fake_services_dict)
+        # Versionless endpoint, config wins
+        config_dict['image_api_version'] = '6'
+        cc = cloud_config.CloudConfig(
+            "test1", "region-al", config_dict, auth_plugin=mock.Mock())
+        cc.get_legacy_client('image', mock_client, version='beef')
+        mock_client.assert_called_with(
+            'beef',
+            service_name=None,
+            endpoint_override='http://example.com',
             region_name='region-al',
             interface='public',
             session=mock.ANY,
@@ -316,6 +414,7 @@ class TestCloudConfig(base.TestCase):
         mock_client.assert_called_with(
             '2.0',
             endpoint_type='public',
+            endpoint_override=None,
             region_name='region-al',
             service_type='network',
             session=mock.ANY,
@@ -333,6 +432,7 @@ class TestCloudConfig(base.TestCase):
         mock_client.assert_called_with(
             '2',
             endpoint_type='public',
+            endpoint_override='http://compute.example.com',
             region_name='region-al',
             service_type='compute',
             session=mock.ANY,
@@ -351,6 +451,7 @@ class TestCloudConfig(base.TestCase):
             '2.0',
             endpoint='http://example.com/v2',
             endpoint_type='admin',
+            endpoint_override=None,
             region_name='region-al',
             service_type='identity',
             session=mock.ANY,
@@ -370,6 +471,7 @@ class TestCloudConfig(base.TestCase):
             '3',
             endpoint='http://example.com',
             endpoint_type='admin',
+            endpoint_override=None,
             region_name='region-al',
             service_type='identity',
             session=mock.ANY,
